# POV Prizm Mosaic, triangle pattern
# Program for conversion of image into a C3 symmetry set of tightly packed triangle prizms,
# colored according to source image pixels, scaled upward according to source image brightness/random
# for artistic appearance, similar to height field
# (c) Ilya Razmanov (mailto:ilyarazmanov@gmail.com)
#
# Input: PNG
# Output: POVRay
#
# History:
# 2007  Initial AmphiSoft POV Sphere Mosaic, using FilterMeister https://filtermeister.com/
# 2023  Rewritten to Python
# 2024  Replaced Pillow I/O with PyPNG from: https://gitlab.com/drj11/pypng
# 2024  Complete internal rewriting. Versions from now on:
#
# 01.000    p3zaika.py Initial release
#
#       Project mirrors:
#       https://github.com/Dnyarri/POVmosaic
#       https://gitflic.ru/project/dnyarri/povmosaic
#

import time
import random
import png
from tkinter import filedialog

# Open source image
sourcefilename = filedialog.askopenfilename(title='Open source PNG file', filetypes=[('PNG','.png')], defaultextension = ('PNG','.png'))
if (sourcefilename == ''):
    quit()

source = png.Reader(filename = sourcefilename)  # starting PyPNG

X,Y,pixels,info = source.asRGBA() # Opening image, iDAT comes to "pixels" as bytearray, to be tuple'd later

totalpixels = X*Y               # Total number OF pixels, not PIXEL NUMBER
rowlength = X*(info['planes'])  # Row length
Z = (info['planes'])            # Maximum CHANNEL NUMBER
imagedata = tuple((pixels))     # Attempt to fix all bytearrays as tuple

if (info['bitdepth'] == 8):
    maxcolors = 255             # Maximal value for 8-bit channel
if (info['bitdepth'] == 16):
    maxcolors = 65535           # Maximal value for 16-bit channel

# Open export file
resultfile = filedialog.asksaveasfile(mode='w', title='Save resulting POV file', filetypes = 
            [
			('POV-Ray scene file', '*.pov'),
            ('All Files', '*.*'),],
    defaultextension = ('POV-Ray scene file','.pov'))
if (resultfile == ''):
    quit()
# Both files opened

# src a-la FM style src(x,y,z)
# Image should be opened as "imagedata" by main program before
# Note that X, Y, Z are not determined in function, you have to determine it in main program

def src(x, y, z):  # Analog src from FM, force repeate edge instead of out of range

    cx = int(x); cy = int(y)
    cx = max(0,cx); cx = min((X-1),cx)
    cy = max(0,cy); cy = min((Y-1),cy)

    position = (cx*Z) + z   # Here is the main magic of turning two x, z into one array position
    channelvalue = int(((imagedata[cy])[position]))
    
    return channelvalue
# end of src function

def srcY(x, y):  # Converting to greyscale, returns Y, force repeate edge instead of out of range

    cx = int(x); cy = int(y)
    cx = max(0,cx); cx = min((X-1),cx)
    cy = max(0,cy); cy = min((Y-1),cy)

    if (info['planes'] < 3):    # supposedly L and LA
        Yntensity = src(x, y, 0)
    else:                       # supposedly RGB and RGBA
        Yntensity = int(0.2989*src(x, y, 0) + 0.587*src(x, y, 1) + 0.114*src(x, y, 2))
    
    return Yntensity
# end of srcY function

#	WRITING POV FILE

# ------------
#  POV header
# ------------

resultfile.write('// Persistence of Vision Ray Tracer Scene Description File\n')
resultfile.write('// Vers: 3.7\n')
resultfile.write('// Description: Mosaic picture consisting from solid triangle prisms, C3 symmetry packing\n')
resultfile.write('// Auth: Automatically generated by p3zaika program, based on AmphiSoft POV Sphere Mosaic plug-in\n\n')
resultfile.write('// developed by Ilya Razmanov\n// mailto:ilyarazmanov@gmail.com\n\n')
resultfile.write('// Project mirrors:\n')
resultfile.write('// https://github.com/Dnyarri/POVmosaic\n')
resultfile.write('// https://gitflic.ru/project/dnyarri/povmosaic\n\n')
resultfile.write(f'// Converted from: {sourcefilename} ')
seconds = time.time()
localtime = time.ctime(seconds)
resultfile.write(f'at: {localtime}\n')
resultfile.write(f'// Source info: {info}\n\n')

#  Statements

resultfile.write('#version 3.7;\n\n')
resultfile.write('global_settings\n')
resultfile.write('{\n')
resultfile.write('  max_trace_level 3\n')
resultfile.write('  adc_bailout 0.01\n')
resultfile.write('  ambient_light <0.5, 0.5, 0.5>\n')
resultfile.write('  assumed_gamma 1.0\n')
resultfile.write('}\n\n')

# Standard includes
resultfile.write('#include "finish.inc"\n#include "golds.inc"\n#include "metals.inc"\n\n')

# Thingie element
resultfile.write('\n// Object thingie\n')
resultfile.write('#declare sqrtof3 = 1.7320508075688772935274463415059;       // sqrt(3)\n')
resultfile.write('#declare sqrtof3div2 = 0.86602540378443864676372317075294;   // sqrt(3)/2\n')
resultfile.write('#declare thingie = prism {\n    linear_sweep\n    linear_spline\n    0,\n    1,\n    4,\n    <-1.0, sqrtof3div2>, <1.0, sqrtof3div2>, <0, -sqrtof3div2>, <-1.0, sqrtof3div2>\n    rotate x*90\n}\n') # Trigonal prism 2.0 wide sqrtof3 tall
resultfile.write('#declare thingie_finish = finish{ambient 0.1 diffuse 0.7 specular 0.8 roughness 0.001}\n')
resultfile.write('#declare color_factor = 1.5;   // Color multiplier for all channels\n')
resultfile.write('#declare displace_factor = 0.5;   // z-Displace multiplier for all thingies\n')
resultfile.write('#declare xyzsize = 1.0;  // x,y,z-Size value for all thingies, Default 1.0\n')
resultfile.write('#declare zmultiplier = 1.0;  // z-Size multiplier for all thingies rescaling, Default 1.0\n')

resultfile.write('\n// Thingie properties - normal\n')
resultfile.write('#declare normalheight = 0.5;  // Normal intensity. Default 0.5\n')
resultfile.write('#declare normalangle = 45;  // Normal is stretched at this angle to X, deg\n')
resultfile.write('#declare normalanglerange = 45;  //Normal stretching angle randomy varies within this range, deg\n')
resultfile.write(f'#declare normalrand = seed({int(seconds*10000000)});  // Seeding random\n')   # will rotate normal randomly around z

# Object "thething" made of thingies

resultfile.write('\n// Object thething made out of thingies\n')
resultfile.write('#declare thething = union {\n')  # Opening object "thething"

# Internal strings for packing change

thingieheight = 1.7320508075688772935274463415059 # Height of tringie with 2*1 sides. sqrt(3) = 1.732... hardcoded to remove math export

translatestring = ' '
oddtranslatestring = ' '  # no offset
eventranslatestring = ' translate <1.0, 0, 0> ' # 1.0 x offset

# Now going to cycle through image and build object
Ycount = int(Y/thingieheight)

for y in range(0, Ycount, 1):

    resultfile.write(f'\n\n // Row {y}\n')

    if (((y+1)%2)==0):
        translatestring = eventranslatestring
    else:
        translatestring = oddtranslatestring

    for x in range(0, X, 1):

        r = float(src(x,y*thingieheight,0))/maxcolors; g = float(src(x,y*thingieheight,1))/maxcolors; b = float(src(x,y*thingieheight,2))/maxcolors       # Normalize colors to 0..1.0
        a = float(src(x,y,3))/maxcolors     # a = 0 - transparent, a = 1.0 - opaque
        yarkost = float(0.2989*r)+float(0.587*g)+float(0.114*b) # brightness
        tobeornottobe = random.random()     # to be used for alpha dithering
        zdisplacement = 0.0                 # original thingie position
        # Three strings below may be commented out to remove brightness dependence of z-displacement but NOT scaling
        zdisplacement = random.random()   # to be used for thingie z-displacement
        # zdisplacement = yarkost           # alternative thingie z-displacement
        # zdisplacement = yarkost * random.random()     # yet alternative thingie z-displacement
        zsize = 1.0                         # original thingie size
        # Two strings below may be commented out to remove brightness dependence of scaling but NOT z-displacement
        zsize = yarkost * zsize             # thingie z-scaling according to source brighness
        zsize = zsize - 0.05*yarkost*random.random()     # further adding random to z-scaling, keeping 0..1 range
        # Next string may not be commented out since it prevent divide by zero in POVRay
        zsize = max(zsize, 1.0/maxcolors)           # Important to avoid divide by zero

        if (a > tobeornottobe):           # whether to draw thingie in place of partially transparent pixel or not
            resultfile.write('  object {thingie pigment {')     # Opening object "thingie" to draw
            resultfile.write(f'rgb <color_factor*{r}, color_factor*{g}, color_factor*{b}>')
            resultfile.write('} finish {thingie_finish} normal {bumps normalheight scale <0.5, 0.05, 0.1> rotate z*(normalangle + normalanglerange*rand(normalrand) - 0.5*normalanglerange)}')   # per-object normal, based on POVRay rand.
            # resultfile.write('} finish {thingie_finish}')    # legacy close finish without normal
            resultfile.write(f' scale <xyzsize,xyzsize,zmultiplier*xyzsize*{zsize}>')        
            if (((x+1)%2)==0):
                resultfile.write(' scale <1.0, -1.0, 1.0>')     # Flipping thingies along the row
            resultfile.write(translatestring)
            resultfile.write(f'translate <{x}, {y*thingieheight}, displace_factor*{zdisplacement}>')
            resultfile.write('}\n')                 # Closing object "thingie" after modifications

# Transform object to fit 1, 1, 1 cube at 0, 0, 0 coordinates
resultfile.write('\n// Object transforms to fit 1, 1, 1 cube at 0, 0, 0 coordinates\n')
resultfile.write('translate <1.0, sqrtof3, 0>\n')   # compensate for extra, now object fit 0..X, 0..Y, 0..maxcolors
resultfile.write(f'translate <-0.5*{X+2}, -0.5*{Y+0.5*thingieheight}, 0>\n')  # translate to center object bottom at x = 0, y = 0, z = 0
resultfile.write(f'scale <-1.0/{max(X,Y)+2}, -1.0/{max(X,Y)+2}, 1.0/{max(X,Y)+2}>\n')  # rescale, mirroring POV coordinates to match Photoshop coordinate system

resultfile.write('} // thething closed\n\n')   # Closing object "thething"

# Insert object into scene
resultfile.write('object {thething}     // good place to change whole object properties\n\n')

# Camera
proportions = max(X,Y)/X
resultfile.write('#declare camera_height = 3.0;\n\n')
resultfile.write('camera {\n   // orthographic\n    location <0.0, 0.0, camera_height>\n    right x*image_width/image_height\n    up y\n    direction <0,0,1>\n    angle 2.0*(degrees(atan2(')
resultfile.write(f'{0.5 * proportions}')
resultfile.write(f', camera_height-(1.0/{max(X,Y)})))) // Supposed to fit object \n    look_at <0.0, 0.0, 0.0>')
resultfile.write('\n}\n\n')

# Light 1
resultfile.write('light_source {0*x\n   color rgb <1.1,1,1>\n   translate <4, 2, 3>}\n\n')
# Light 2
resultfile.write('/* light_source {0*x\n   color rgb <0.9,1,1>\n   translate <-2, 6, 7>} */\n\n')
resultfile.write('\n/*\n\nhappy rendering\n\n  0~0\n (---)\n(.>|<.)\n-------\n\n*/')
# Close output
resultfile.close()